#include "chr_cursor.h"

#include <stdint.h>

chr_cursor::chr_cursor(chr_table * table, size_t key_size,
    size_t value_size)
  : key_size(key_size), value_size(value_size), table(table),
  leaf(NULL), curkey(NULL), pos(CHR_UNKNOWN_POS)
{
  curkey = (unsigned char *)malloc(key_size);
  if(curkey == NULL)
  {
    throw std::bad_alloc();
  }
  prevkey = (unsigned char *)malloc(key_size);
  memset(prevkey, 0, key_size);
}

chr_cursor::~chr_cursor()
{
  leaf->unbind();
  free(curkey);
  free(prevkey);
}

/* returns CHR_OK */
int chr_cursor::move(const unsigned char * key, chr_pos_type type,
    chr_pos_where where)
{
  assert(type != CHR_KEY || key != NULL);
  if(leaf != NULL)
  {
    leaf->unbind();
    leaf = NULL;
  }
  this->where = where;
  leaf = table->search(key, type);
  leaf->bind();
  pos = CHR_UNKNOWN_POS;
  if(type == CHR_FIRST)
  {
    unsigned char * dummy_val = (unsigned char *)malloc(value_size);
    assert(dummy_val != NULL);
    pos = leaf->read_first(curkey, dummy_val);
    free(dummy_val);
  }
  else if(type == CHR_LAST)
  {
    unsigned char * dummy_val = (unsigned char *)malloc(value_size);
    assert(dummy_val != NULL);
    pos = leaf->read_last(curkey, dummy_val);
    free(dummy_val);
  }
  else
  {
    assert(type == CHR_KEY);
    assert(key != NULL);
    memcpy(curkey, key, key_size);
  }
  table->print();
  return(CHR_OK);
}

/* returns CHR_OK or CHR_DUPKEY */
int chr_cursor::insert(const unsigned char * key, const unsigned char * val)
{
  assert(key != NULL);
  assert(val != NULL);
  if(leaf == NULL)
  {
    fprintf(stderr, "error: cursor not set\n");
    return(CHR_ERR);
  }
  int ret = leaf->insert(key, val);
  if(ret == CHR_OOR)
  {
    set_leaf(table->search(key, CHR_KEY));
    ret = leaf->insert(key, val);
    assert(ret != CHR_OOR);
  }
  /* not useful for consecutive insert: disabled
     can be used with mixed insert/read patterns though,
     whose behavior is currently unexpected:
     insert does not move the cursor */
  /*
    memcpy(curkey, key, key_size);
    where = CHR_AFTER;
  */
  /*
  table->print();
  table->dbug_check();
  */
  return(ret);
}

/* returns CHR_OK or CHR_NOTFOUND */
int chr_cursor::del(const unsigned char * key)
{
  assert(key != NULL);
  if(leaf == NULL)
  {
    fprintf(stderr, "error: cursor not set\n");
    return(CHR_ERR);
  }
  int ret = leaf->del(key);
  if(ret == CHR_OOR)
  {
    set_leaf(table->search(key, CHR_KEY));
    ret = leaf->del(key);
    assert(ret != CHR_OOR);
  }
  table->print();
  return(ret);
}

/* returns CHR_OK or CHR_NOTFOUND */
int chr_cursor::update(const unsigned char * key, const unsigned char * val)
{
  assert(key != NULL);
  assert(val != NULL);
  if(leaf == NULL)
  {
    fprintf(stderr, "error: cursor not set\n");
    return(CHR_ERR);
  }
  int ret = leaf->update(key, val);
  if(ret == CHR_OOR)
  {
    set_leaf(table->search(key, CHR_KEY));
    ret = leaf->update(key, val);
    assert(ret != CHR_OOR);
  }
  table->print();
  return(ret);
}

int chr_cursor::read_next(unsigned char * key, unsigned char * val)
{
  assert(key != NULL);
  assert(val != NULL);
  if(leaf == NULL)
  {
    fprintf(stderr, "error: cursor not set\n");
    return(CHR_ERR);
  }
  if(where == CHR_BEFORE)
  {
    pos = leaf->read_exactornext(curkey, val);
    assert(pos != CHR_NEXTLEAF);
    /* NEXTLEAF is not possible for exactornext, ensure that */
    if(pos == CHR_OOR)
    {
      set_leaf(table->search(curkey, CHR_KEY));
      pos = leaf->read_exactornext(curkey, val);
      assert(pos != CHR_OOR && pos != CHR_NEXTLEAF);
    }
    if (pos == CHR_NOTFOUND)
    {
      return(CHR_EOF);
    }
    memcpy(key, curkey, key_size);
    where = CHR_ON;
    return(CHR_OK);
  }
  /* else: CHR_ON or CHR_AFTER */
  assert(where == CHR_ON || where == CHR_AFTER);
  pos = leaf->readnext(curkey, key, val, pos);
  if(pos == CHR_OOR)
  {
    set_leaf(table->search(curkey, CHR_KEY));
    pos = leaf->readnext(curkey, key, val, -1);
    assert(pos != CHR_OOR);
  }
  if(pos == CHR_NEXTLEAF)
  {
    if(leaf->parent == NULL)
    {
      return(CHR_EOF);
    }
    chr_inode * parent = leaf->parent;
    chr_node * son;
    while((son = parent->nextson(curkey)) == NULL)
    {
      parent = parent->parent;
      if(parent == NULL)
      {
        return(CHR_EOF);
      }
    }
    while(!son->is_leaf())
    {
      parent = son->inode();
      son = parent->son(curkey);
    }
    set_leaf(son->leaf());
    pos = leaf->readnext(curkey, key, val, -1);
    assert(pos != CHR_NEXTLEAF);
    assert(pos != CHR_OOR);
  }
  memcpy(curkey, key, key_size);
  return(CHR_OK);
}

int chr_cursor::read_prev(unsigned char * key, unsigned char * val)
{
  assert(key != NULL);
  assert(val != NULL);
  if(leaf == NULL)
  {
    fprintf(stderr, "error: cursor not set\n");
    return(CHR_ERR);
  }
  if(where == CHR_AFTER)
  {
    pos = leaf->read_exactorprev(curkey, val);
    assert(pos != CHR_NEXTLEAF);
    /* NEXTLEAF is not possible for exactorprev, ensure that */
    if(pos == CHR_OOR)
    {
      set_leaf(table->search(curkey, CHR_KEY));
      pos = leaf->read_exactorprev(curkey, val);
      assert(pos != CHR_OOR && pos != CHR_NEXTLEAF);
    }
    if (pos == CHR_NOTFOUND)
    {
      return(CHR_EOF);
    }
    memcpy(key, curkey, key_size);
    where = CHR_ON;
    return(CHR_OK);
  }
  /* else: CHR_ON or CHR_BEFORE */
  assert(where == CHR_ON || where == CHR_BEFORE);
  pos = leaf->readprev(curkey, key, val, pos);
  if(pos == CHR_OOR)
  {
    set_leaf(table->search(curkey, CHR_KEY));
    pos = leaf->readprev(curkey, key, val, -1);
    assert(pos != CHR_OOR);
  }
  if(pos == CHR_PREVLEAF)
  {
    if(leaf->parent == NULL)
    {
      return(CHR_EOF);
    }
    chr_inode * parent = leaf->parent;
    chr_node * son;
    while((son = parent->prevson(curkey)) == NULL)
    {
      parent = parent->parent;
      if(parent == NULL)
      {
        return(CHR_EOF);
      }
    }
    while(!son->is_leaf())
    {
      parent = son->inode();
      son = parent->son(curkey);
    }
    set_leaf(son->leaf());
    pos = leaf->readprev(curkey, key, val, -1);
    assert(pos != CHR_PREVLEAF);
    assert(pos != CHR_OOR);
  }
  memcpy(curkey, key, key_size);
  return(CHR_OK);
}

inline void chr_cursor::set_leaf(chr_leaf * leaf)
{
  assert(leaf != NULL);
  assert(this->leaf != NULL);
  leaf->bind();
  this->leaf->unbind();
  this->leaf = leaf;
}

