#include "chr_common.h"

#include "chr_flash.h"

#include <assert.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>


/* the size used by the database should be set to a given value,
 * or should be determined by a percentage of free sectors */
chr_flash::chr_flash(const char * device, const char * desc_file,
    chr_flashgui * gui)
  : gui(gui)
{
  pthread_mutex_init(&mutex, NULL);
  free_sectors = new chr_list<chr_lsn, NULL_LSN>();
  this->desc_file = (char*)malloc(strlen(desc_file) + 1);
  assert(this->desc_file != NULL);
  strcpy(this->desc_file, desc_file);
  FILE * desc_fd = fopen(desc_file, "r");
  uint64_t nb_free_sectors;
  if(desc_fd == NULL ||
      fread(&nb_free_sectors, sizeof(uint64_t), 1, desc_fd) != 1)
  {
    /* init free sectors list */
    max_lsn = CHR_DEV_SIZE_INCR;
    gui->set_size(max_lsn);
    for(chr_lsn i = 1; i < CHR_DEV_SIZE_INCR; i++)
    {
      gui->set_free(i);
      free_sectors->insert(i);
    }
  }
  else
  {
    /* load free sectors list */
    if(fread(&max_lsn, sizeof(chr_lsn), 1, desc_fd) != 1 ||
        max_lsn < nb_free_sectors)
    { 
      fprintf(stderr, "error: %s is corrupted\n", desc_file);
    }
    gui->set_size(max_lsn);
    chr_lsn lsn;
    for(uint64_t i = 0; i < nb_free_sectors; i++)
    {
      if(fread(&lsn, sizeof(chr_lsn), 1, desc_fd) != 1)
      {
        fprintf(stderr, "error: %s is corrupted\n", desc_file);
        i = nb_free_sectors;
      }
      assert(lsn != NULL_LSN);
      gui->set_free(lsn);
      free_sectors->insert(lsn);
    }
    fclose(desc_fd);
  }
  fd = open(device, O_RDWR | O_LARGEFILE,
      S_IRWXU | S_IRWXG | S_IRWXO);
  if (fd == -1)
  {
    fprintf(stderr, "warning: flash device does not exist, attempt to create %s\n", device);
    fd = open(device, O_RDWR | O_LARGEFILE | O_CREAT,
        S_IRWXU | S_IRWXG | S_IRWXO);
    if(fd == -1)
    {
      fprintf(stderr, "error: can not create flash device %s\n", device);
      exit(EXIT_FAILURE);
    }
  }
}

chr_flash::~chr_flash()
{
  pthread_mutex_lock(&mutex);
  FILE * desc_fd = fopen(desc_file, "w+");
  if(desc_fd == NULL)
  {
    fprintf(stderr, "error: can not open %s\n", desc_file);
    fprintf(stderr, "free sectors will not be saved\n");
  }
  else
  {
    /* save free sectors list */
    uint64_t nb_free_sectors = 0;
    fwrite(&nb_free_sectors, sizeof(uint64_t), 1, desc_fd);
    fwrite(&max_lsn, sizeof(chr_lsn), 1, desc_fd);
    chr_lsn lsn = free_sectors->pop();
    while(lsn != NULL_LSN)
    {
      fwrite(&lsn, sizeof(chr_lsn), 1, desc_fd);
      nb_free_sectors++;
      lsn = free_sectors->pop();
    }
    fseek(desc_fd, 0, SEEK_SET);
    fwrite(&nb_free_sectors, sizeof(uint64_t), 1, desc_fd);
    fclose(desc_fd);
  }
  pthread_mutex_destroy(&mutex);
  close(fd);
  free(desc_file);
  delete free_sectors;
}

int chr_flash::read(const chr_lsn lsn, unsigned char * buf) const
{
  assert(lsn != NULL_LSN);
  pthread_mutex_lock(&mutex);
  gui->set_reading(lsn);
  ssize_t siz = pread64(fd, buf, CHR_BLOCK_SIZE,
      (off64_t)lsn * CHR_BLOCK_SIZE);
  gui->unset_reading(lsn);
  if(siz != CHR_BLOCK_SIZE)
  {
    fprintf(stderr, "read(%lld) failed (%d), returned %Zd\n",
        lsn, errno, siz);
    pthread_mutex_unlock(&mutex);
    return(CHR_READ_ERR);
  }
  pthread_mutex_unlock(&mutex);
  return(CHR_OK);
}

int chr_flash::nameless_write(chr_lsn * lsn, const unsigned char * buf)
{
  pthread_mutex_lock(&mutex);
  if(free_sectors->size() < (max_lsn * CHR_DEV_THRESHOLD))
  {
    chr_lsn i = max_lsn;
    max_lsn += CHR_DEV_SIZE_INCR;
    gui->set_size(max_lsn);
    while(i < max_lsn)
    {
      free_sectors->insert(i);
      i++;
    }
  }
  *lsn = free_sectors->pop();
  if(*lsn == NULL_LSN)
  {
    fprintf(stderr, "not enough space on flash memory\n");
    pthread_mutex_unlock(&mutex);
    return(CHR_WRITE_ERR);
  }
  gui->set_used(*lsn);
  gui->set_nl_writing(*lsn);
  ssize_t siz = pwrite64(fd, buf, CHR_BLOCK_SIZE,
      (off64_t)(*lsn) * CHR_BLOCK_SIZE);
  gui->unset_nl_writing(*lsn);
  if(siz != CHR_BLOCK_SIZE)
  {
    fprintf(stderr, "nameless_write(%lld) failed (%d), returned %Zd\n",
        *lsn, errno, siz);
    pthread_mutex_unlock(&mutex);
    return(CHR_WRITE_ERR);
  }
  pthread_mutex_unlock(&mutex);
  return(CHR_OK);
}

int chr_flash::named_write(const chr_lsn lsn, const unsigned char * buf)
{
  assert(lsn != NULL_LSN);
  pthread_mutex_lock(&mutex);
  /* do not overwrite with nameless writes */
  free_sectors->remove(lsn);
  gui->set_used(lsn);
  gui->set_writing(lsn);
  ssize_t siz = pwrite64(fd, buf, CHR_BLOCK_SIZE,
      (off64_t)lsn * CHR_BLOCK_SIZE);
  gui->unset_writing(lsn);
  if(siz != CHR_BLOCK_SIZE)
  {
    fprintf(stderr, "named_write(%lld) failed (%d), returned %Zd\n",
        lsn, errno, siz);
    pthread_mutex_unlock(&mutex);
    return(CHR_WRITE_ERR);
  }
  pthread_mutex_unlock(&mutex);
  return(CHR_OK);
}

void chr_flash::trim(const chr_lsn lsn)
{
  pthread_mutex_lock(&mutex);
  gui->set_trimmed(lsn);
  free_sectors->insert(lsn);
  gui->set_free(lsn);
  pthread_mutex_unlock(&mutex);
}

