#include "chr_leaf.h"

#include "chr_parse.h"

chr_leaf::chr_leaf(size_t key_size, size_t value_size,
    chr_inode * parent, chr_table * table, chr_flash * flash,
    bool kmininf, const unsigned char * kmin, bool kmaxinf,
    const unsigned char * kmax, chr_lsn lsn, const unsigned char * buf)
: chr_node(key_size,
    chr_parse::keys_per_leaf_memory(key_size, value_size),
    parent, table, flash, kmininf, kmin, kmaxinf, kmax),
  lsn(lsn),
  value_size(value_size),
  max_key_epsilon(NULL),
  values(NULL),
  buf(NULL),
  dirty(false)
{
  assert(keys != NULL);
  this->max_key_epsilon = (unsigned char *)malloc(key_size);
  this->values = (unsigned char *)malloc(value_size * keys_per_node);
  if(this->values == NULL ||
      this->max_key_epsilon == NULL)
  {
    free(this->max_key_epsilon);
    free(this->values);
    throw std::bad_alloc();
  }
  memset(max_key_epsilon, 0, key_size);

  if(lsn == NULL_LSN) /* empty (new) leaf */
  {
    nb_keys = 0;
  }
  else
  {
    /* load from disk */
    assert(buf != NULL);
    nb_keys = chr_parse::leaf_nb_keys(buf);
    for(unsigned int i = 0; i<nb_keys; i++)
    {
      memcpy(key(i), chr_parse::leaf_key(buf, i, key_size, value_size),
          key_size);
      memcpy(value(i), chr_parse::
          leaf_value(buf, i, key_size, value_size), value_size);
    }
  }
  if(parent != NULL)
  {
    parent->bind();
  }
}

chr_leaf::~chr_leaf()
{
  if(dirty)
  {
    /* flush tuples if any */
    chr_lsn old_lsn = lsn;
    if(nb_keys > 0)
    {
      if(buf == NULL)
      {
        buf = (unsigned char *)malloc(CHR_BLOCK_SIZE);
      }
      if(buf == NULL)
      {
        fprintf(stderr, "error: not enough memory to write leaf, "
            "data might be lost\n");
      }
      else
      {
        this->to_sector(buf);
        flash->nameless_write(&lsn, buf);
      }
    }
    if(old_lsn != NULL_LSN)
    {
      /* free previous sector */
      flash->trim(old_lsn);
    }
  }
  if(nb_keys > 0)
  {
    if(parent != NULL)
    {
      parent->update(key(0), lsn);
    }
    else
    {
      table->root = NULL;
      table->root_lsn = lsn;
    }
  }
  else
  {
    if(parent != NULL)
    {
      parent->del(kmin);
    }
    else
    {
      table->root = NULL;
      table->root_lsn = NULL_LSN;
    }
  }
  free(max_key_epsilon);
  free(values);
  free(buf);
  if(parent != NULL)
  {
    parent->unbind();
  }
  else
  {
    table->root = NULL;
  }
}

void chr_leaf::to_sector(unsigned char * buf)
{
  while(nb_keys > chr_parse::keys_per_leaf_flash(key_size, value_size))
  {
    /* split in the middle */
    memcpy(max_key_epsilon, key(nb_keys/2), key_size);
    split();
  }
  assert(nb_keys <= chr_parse::keys_per_leaf_flash(key_size, value_size));
  chr_parse::leaf_to_sector(
      buf, nb_keys, key_size, value_size, keys, values);
}

int chr_leaf::split()
{
  assert(parent == NULL || parent->son(key(0)) == this);
  if(parent == NULL)
  {
    /* allocate new root */
    parent = new chr_inode(key_size, NULL, table, flash,
        true, NULL, true, NULL, NULL_LSN, NULL);
    assert(parent != NULL);
    parent->init_son(this);
    /* bind twice, for this node and for being root */
    parent->bind();
    parent->bind();
    /* unbind previous root (this) */
    this->unbind();
    table->root = parent;
  }

  /* split leaf */
  if(buf == NULL)
  {
    buf = (unsigned char *)malloc(CHR_BLOCK_SIZE);
    assert(buf != NULL);
  }
  unsigned int keys_per_sector = chr_parse::
    keys_per_leaf_flash(key_size, value_size);
  assert(nb_keys > keys_per_sector);
  if(comp(max_key_epsilon, key(keys_per_sector - 1)) < 0)
  {
    /* worst case:
     * associate last keys with a new node to be flushed on disk */

    /* split key should be key(max_pos_epsilon + 1) */

    /* if epsilon is large, we might have to write multiple sectors.
     * A workaround is to write only the last m tuples
     * when that happens */
    int max_pos_epsilon;
    search(&max_pos_epsilon, max_key_epsilon);
    max_pos_epsilon++;
    if(nb_keys - max_pos_epsilon >
        chr_parse::keys_per_leaf_flash(key_size, value_size))
    {
      max_pos_epsilon =
        nb_keys - chr_parse::keys_per_leaf_flash(key_size, value_size);
    }
    set_kmax(false, key(max_pos_epsilon));

    /* new_leaf (buf) <- last e keys >= split_key */
    chr_parse::leaf_to_sector(buf, nb_keys - max_pos_epsilon,
        key_size, value_size, key(max_pos_epsilon),
        value(max_pos_epsilon));

    /* remove last e keys from leaf */
    nb_keys = max_pos_epsilon;
    max_pos_epsilon = 0;
    memset(max_key_epsilon, 0, key_size);

    chr_lsn new_leaf_lsn;
    flash->nameless_write(&new_leaf_lsn, buf);
    return(parent->insert_right(kmax, new_leaf_lsn));
  }
  else
  {
    /* best case: flush first keys */
    /* split key is key(keys_per_sector) */
    /* new_leaf (buf) <- first m keys */
    chr_parse::leaf_to_sector(buf, keys_per_sector,
        key_size, value_size, keys, values);

    /* remove first m keys from leaf */
    memmove(keys, key(keys_per_sector),
        (nb_keys - keys_per_sector) * key_size);
    memmove(values, value(keys_per_sector),
        (nb_keys - keys_per_sector) * value_size);
    nb_keys -= keys_per_sector;

    memset(max_key_epsilon, 0, key_size);
    /* split key is now key(0) */
    set_kmin(false, key(0));

    chr_lsn new_leaf_lsn;
    flash->nameless_write(&new_leaf_lsn, buf);
    return(parent->insert_left(kmin, new_leaf_lsn));
  }
}

int chr_leaf::insert(const unsigned char * k, const unsigned char * val)
{
  assert(parent == NULL || parent->son(key(0)) == this);
  assert(nb_keys < keys_per_node);
  if(!in_range(k))
  {
    return(CHR_OOR);
  }
  int pos = nb_keys - 1;
  while(pos >= 0 && comp(key(pos), k) > 0)
  {
    memcpy(key(pos + 1), key(pos), key_size);
    memcpy(value(pos + 1), value(pos), value_size);
    pos--;
  }
  if(pos >= 0 && comp(key(pos), k) == 0)
  {
    /* duplicate key, rollback */
    pos++;
    while((unsigned int)pos < nb_keys)
    {
      memcpy(key(pos), key(pos + 1), key_size);
      memcpy(value(pos), value(pos + 1), value_size);
      pos++;
    }
    return(CHR_DUPKEY);
  }
  pos++;
  memcpy(key(pos), k, key_size);
  memcpy(value(pos), val, value_size);
  nb_keys++;
  if(nb_keys >= keys_per_node - CHR_EPSILON &&
      comp(k, max_key_epsilon) > 0)
  {
    memcpy(max_key_epsilon, k, key_size);
  }
  dirty = true;

  if(nb_keys == keys_per_node)
  {
    split();
  }
  return(CHR_OK);
}

int chr_leaf::update(const unsigned char * k, const unsigned char * val)
{
  assert(parent == NULL || parent->son(key(0)) == this);
  if(!in_range(k))
  {
    return(CHR_OOR);
  }
  int pos;
  int found = search(&pos, k);
  if(found == CHR_NOTFOUND)
  {
    return(CHR_NOTFOUND);
  }
  dirty = true;
  assert(comp(k, key(pos)) == 0);
  memcpy(value(pos), val, value_size);
  return(CHR_OK);
}

int chr_leaf::del(const unsigned char * k)
{
  assert(parent == NULL || parent->son(key(0)) == this);
  if(!in_range(k))
  {
    return(CHR_OOR);
  }
  int pos;
  int found = search(&pos, k);
  if(found == CHR_NOTFOUND)
  {
    return(CHR_NOTFOUND);
  }
  dirty = true;
  assert(comp(k, key(pos)) == 0);
  nb_keys--;
  for(unsigned int i = pos; i < nb_keys; i++)
  {
    memcpy(key(i), key(i+1), key_size);
    memcpy(value(i), value(i+1), value_size);
  }

  return(CHR_OK);
}

int chr_leaf::readnext(const unsigned char * prevkey, unsigned char * k,
    unsigned char * val, int pos_hint)
{
  assert(parent == NULL || parent->son(key(0)) == this);
  assert(pos_hint >= -1 || pos_hint == CHR_UNKNOWN_POS);
  if(dirty || pos_hint == CHR_UNKNOWN_POS)
  {
    /* if(!in_range(prevkey)) */
    /* verify that prevkey <= kmax */
    if (!kmaxinf && comp(prevkey, kmax) > 0)
    {
      return(CHR_OOR);
    }
    /* verify hint, could be more efficient */
    search(&pos_hint, prevkey);
  }
  assert(pos_hint < (int)nb_keys);
  assert(pos_hint >= -1);
  assert(pos_hint == -1 ||
      comp(key(pos_hint), prevkey) == 0 ||
      comp(key(pos_hint), prevkey) < 0);
  pos_hint++;
  if((unsigned int)pos_hint == nb_keys)
  {
    return(CHR_NEXTLEAF);
  }
  assert(comp(key(pos_hint), prevkey) > 0);
  memcpy(k, key(pos_hint), key_size);
  memcpy(val, value(pos_hint), value_size); 

  return(pos_hint);
}

int chr_leaf::readprev(const unsigned char * prevkey, unsigned char * k,
    unsigned char * val, int pos_hint)
{
  assert(parent == NULL || parent->son(key(0)) == this);
  assert(pos_hint >= -1 || pos_hint == CHR_UNKNOWN_POS);
  if(dirty || pos_hint == CHR_UNKNOWN_POS)
  {
    /* if(!in_range(prevkey)) */
    /* verify that prevkey >= kmin */
    if (!kmininf && comp(prevkey, kmin) < 0)
    {
      return(CHR_OOR);
    }
    /* verify hint, could be more efficient */
    search(&pos_hint, prevkey);
  }
  assert(pos_hint < (int)nb_keys);
  assert(pos_hint >= -1);
  assert(pos_hint == -1 ||
      comp(key(pos_hint), prevkey) == 0 ||
      comp(key(pos_hint), prevkey) < 0);
  if((unsigned int)pos_hint == -1)
  {
    return(CHR_PREVLEAF);
  }
  assert(comp(key(pos_hint), prevkey) < 0);
  memcpy(k, key(pos_hint), key_size);
  memcpy(val, value(pos_hint), value_size); 

  return(pos_hint);
}

int chr_leaf::read_exact(const unsigned char * k, unsigned char * val)
{
  assert(parent == NULL || parent->son(key(0)) == this);
  if(!in_range(k))
  {
    return(CHR_OOR);
  }
  int pos;
  if(search(&pos, k) == CHR_NOTFOUND)
  {
    return(CHR_NOTFOUND);
  }
  memcpy(val, value(pos), value_size);
  return(pos);
}

int chr_leaf::read_exactornext(unsigned char * k, unsigned char * val)
{
  assert(parent == NULL || parent->son(key(0)) == this);
  if(!in_range(k))
  {
    return(CHR_OOR);
  }
  int pos;
  if(search(&pos, k) == CHR_NOTFOUND)
  {
    pos++;
  }
  if((unsigned int)pos == nb_keys)
  {
    return(CHR_NOTFOUND);
  }
  memcpy(k, key(pos), key_size);
  memcpy(val, value(pos), value_size);
  return(pos);
}

int chr_leaf::read_exactorprev(unsigned char * k, unsigned char * val)
{
  assert(parent == NULL || parent->son(key(0)) == this);
  if(!in_range(k))
  {
    return(CHR_OOR);
  }
  int pos;
  search(&pos, k);
  if(pos == -1)
  {
    return(CHR_NOTFOUND);
  }
  memcpy(k, key(pos), key_size);
  memcpy(val, value(pos), value_size);
  return(pos);
}

int chr_leaf::read_first(unsigned char * k, unsigned char * val)
{
  assert(parent == NULL || parent->son(key(0)) == this);
  if(nb_keys == 0)
  {
    return(CHR_OOR);
  }
  memcpy(k, key(0), key_size);
  memcpy(val, value(0), value_size);
  return(0);
}

int chr_leaf::read_last(unsigned char * k, unsigned char * val)
{
  assert(parent == NULL || parent->son(key(0)) == this);
  if(nb_keys == 0)
  {
    return(CHR_OOR);
  }
  memcpy(k, key(nb_keys - 1), key_size);
  memcpy(val, value(nb_keys - 1), value_size);
  return(nb_keys - 1);
}

unsigned char * chr_leaf::value(unsigned int x)
{
  return(&(values[x * value_size]));
}

bool chr_leaf::is_leaf()
{
  return(true);
}

chr_leaf * chr_leaf::leaf()
{
  return(this);
}

chr_inode * chr_leaf::inode()
{
  /* this is not an inode */
  assert(false);
  return(NULL);
}

void chr_leaf::drop()
{
  if(lsn != NULL_LSN)
  {
    flash->trim(lsn);
  }
  delete this;
}

void chr_leaf::get_state(chr_lsn * lsn, bool * kmininf,
    unsigned char ** kmin, bool * kmaxinf, unsigned char ** kmax,
    size_t * key_size, size_t * value_size, unsigned int * nb_keys,
    unsigned char ** keys, unsigned char ** values)
{
  *lsn = this->lsn;
  *kmininf = this->kmininf;
  *kmin = this->kmin;
  *kmaxinf = this->kmaxinf;
  *kmax = this->kmax;
  *key_size = this->key_size;
  *value_size = this->value_size;
  *nb_keys = this->nb_keys;
  *keys = this->keys;
  *values = this->values;
}

void chr_leaf::dbug_check(bool kmininf, const unsigned char * kmin,
    bool kmaxinf, const unsigned char * kmax)
{
  /* kmin <= ki < ki+1 < kmax */
  if(nb_keys == 0)
  {
    assert(kmininf || kmaxinf || comp(kmin, kmax) < 0);
  }
  else
  {
    assert(kmininf || comp(kmin, key(0)) <= 0);
    assert(kmaxinf || comp(key(nb_keys - 1), kmax) < 0);
    for(unsigned int i = 0; i < nb_keys - 1; i++)
    {
      assert(comp(key(i), key(i + 1)) < 0);
    }
  }
}

