#ifndef CHR_PARSE_H
#define CHR_PARSE_H

#include "chr_common.h"

#include <stdlib.h>
#include <stdint.h>
#include <assert.h>
#include <arpa/inet.h>

#include "chr_ptr.h"

/* assume little-endian system 
inline uint16_t endian_swap16(uint16_t x)
{
  return((x>>8) | (x<<8));
}
*/

/* node layout :
   nb_keys | ptr [ key | ptr ]*
   nb_keys is a 16bit unsigned integer in big endian,
     with its first bit set to 0
   ptr is a 64bit unsigned integer
   key is table-dependant */

/* leaf layout :
 * table_id | nb_keys [ key | value ]*
   table id is a 8 bit unique table identifier,
     with it's first bit set to 1
   nb_keys is a 16bit unsigned integer in big endian,
   key is table-dependant
   value is table-dependant */

namespace chr_parse
{

inline unsigned int keys_per_leaf_flash(size_t key_size, size_t value_size)
{
  return((CHR_BLOCK_SIZE - 3) / (key_size + value_size));
}

inline unsigned int keys_per_leaf_memory(size_t key_size, size_t value_size)
{
  return(keys_per_leaf_flash(key_size, value_size) + CHR_EPSILON);
}

inline unsigned int keys_per_inode(size_t key_size)
{
  return((CHR_BLOCK_SIZE - 2 - sizeof(chr_lsn)) /
      (key_size + sizeof(chr_lsn)));
}

inline bool is_leaf(const unsigned char * buf)
{
  return((buf[0] & (unsigned char)0x80) == (unsigned char)0x80);
}

inline unsigned int leaf_nb_keys(const unsigned char * buf)
{
  uint16_t ret;
  assert(is_leaf(buf));
  memcpy(&ret, buf+1, sizeof(uint16_t));
  return(ntohs(ret));
}

inline unsigned int inode_nb_keys(const unsigned char * buf)
{
  uint16_t ret;
  memcpy(&ret, buf, sizeof(uint16_t));
  assert((ntohs(ret) & 0x8000) == 0x0000);
  return(ntohs(ret));
}

inline const unsigned char * leaf_key(const unsigned char * buf,
    unsigned int i, size_t key_size, size_t value_size)
{
  return(&(buf[3 + i * (key_size + value_size)]));
}

inline const unsigned char * leaf_value(const unsigned char * buf,
    unsigned int i, size_t key_size, size_t value_size)
{
  return(&(buf[3 + i * (key_size + value_size) + key_size]));
}

inline const unsigned char * inode_key(const unsigned char * buf,
    unsigned int i, size_t key_size)
{
  return(&(buf[2 + i * (key_size + sizeof(chr_lsn)) + sizeof(chr_lsn)]));
}

inline const chr_lsn inode_ptr(const unsigned char * buf,
    unsigned int i, size_t key_size)
{
  chr_lsn * ret=(chr_lsn *)(&(buf[2 + i * (key_size + sizeof(chr_lsn))]));
  assert(*ret != NULL_LSN);
  return *ret;
}

inline void leaf_to_sector(unsigned char * buf, unsigned int nb_tuples,
  size_t key_size, size_t value_size, const unsigned char * keys,
  const unsigned char * values)
{
  assert(buf != NULL);
  memset(buf, 0x80, 1);
  unsigned char * parser = buf+1;
  uint16_t nb_keys_be = htons(nb_tuples);
  memcpy(parser, &nb_keys_be, sizeof(uint16_t));
  parser+=sizeof(uint16_t);
  for(unsigned int i = 0; i < nb_tuples; i++)
  {
    memcpy(parser, keys, key_size);
    parser+=key_size;
    keys+=key_size;
    memcpy(parser, values, value_size);
    parser+=value_size;
    values+=value_size;
  }
  memset(parser, 0, CHR_BLOCK_SIZE - (parser - buf));
}

inline void inode_to_sector(unsigned char * buf, unsigned int nb_keys,
    size_t key_size, const unsigned char * keys, const chr_ptr * ptrs)
{
  assert(buf != NULL);
  uint16_t nb_keys_be = htons(nb_keys);
  /* first bit should not be set */
  assert((nb_keys_be & 0x80) == 0x00);
  unsigned char * parser = buf;
  memcpy(parser, &nb_keys_be, sizeof(uint16_t));
  parser+=sizeof(uint16_t);
  chr_lsn tmp_lsn = ptrs->get_lsn();
  assert(tmp_lsn != NULL_LSN);
  memcpy(parser, &tmp_lsn, sizeof(chr_lsn));
  parser+=sizeof(chr_lsn);
  ptrs++;
  for(unsigned int i = 0; i < nb_keys; i++)
  {
    memcpy(parser, keys, key_size);
    parser+=key_size;
    keys+=key_size;
    assert(ptrs->is_memory() == false);
    tmp_lsn = ptrs->get_lsn();
    assert(tmp_lsn != NULL_LSN);
    memcpy(parser, &tmp_lsn, sizeof(chr_lsn));
    parser+=sizeof(chr_lsn);
    ptrs++;
  }
  memset(parser, 0, CHR_BLOCK_SIZE - (parser - buf));
}

}

#endif /* CHR_PARSE_H */

